/*==============================================================================*/
/*  Copyright (C) 2009-2016, Heartland Data inc. All Rights Reserved.           */
/*                                                                              */
/*  Title  :   SDIF Driver                                                      */
/*  EventID:   Insert Extended Event-ID Test Point                              */
/*  AppVer :   DT10 Ver10.XX`                                                  */
/*  FileID :   3aa                                                              */
/*  Version:   1.2                                                              */
/*  Author :   HLDC                                                             */
/*==============================================================================*/

/*==============================================================================*/
/*  Desc:   Header for SDI/F IO                                                 */
/*==============================================================================*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/proc_fs.h>
#include <asm/uaccess.h>
#include <linux/kthread.h>
#include <linux/sched.h>
#include <linux/vmalloc.h>
#include <linux/version.h>

#include <linux/mmc/host.h>
#include <linux/mmc/card.h>
#include <linux/mmc/mmc.h>
#include <linux/mmc/core.h>


/*==============================================================================*/
/*  Macro:  DT_ADD_EVENT_INFO                                                   */
/*  Desc:   Please set 1 when you add Event Value to Test Point.                */
/*==============================================================================*/
#define DT_ADD_TIME_INFO	0

/*==============================================================================*/
/*  Macro:  DT_ADD_CPU_INFO                                                     */
/*  Desc:   Please set 1 when you add CPU ID to Test Point.                     */
/*==============================================================================*/
#define DT_ADD_CPU_INFO		0

/*==============================================================================*/
/*  Desc:   Please set SD CMD                                                   */
/*==============================================================================*/
#define NML_TP_CMD			41	/* Normal TP opcode 0x29 */
#define VAL_TP_CMD			51	/* Variable TP opcode 0x33 */
#define VAL_SIZE_CMD		45	/* Variable Size opcode 0x2d */
#define VAL_DAT_CMD			46	/* Variable Data opcode 0x2e */
#define EVENT_ID_CMD		44	/* Eventinfo TP opcode 0x2c */
#define TIME_INFO_CMD		43	/* Timeinfo TP opcode 0x2b */



#define TP_BusOut		'3'
#define TP_MemoryOutput		'4'
#define	PROC_NAME	"dt10proc"

static struct mmc_host *dt10sd_host = NULL;
static u32 start_time;
static int init = 0;
static DEFINE_SPINLOCK(sdif_lock);

struct DT10_DATA {
	char tp_func;
	unsigned int addr;
	unsigned int dat;
	unsigned int size;
	unsigned char value[256];
};

void dt10_sd_host_get(struct mmc_host *host)
{
	dt10sd_host = host;
}
EXPORT_SYMBOL(dt10_sd_host_get);

static inline void dt10_sd_send_cmd(unsigned char *buff, unsigned int opcode)
{
	unsigned long flags;
	struct mmc_command cmd;

	memset(&cmd, 0, sizeof(struct mmc_command));

	spin_lock_irqsave(&sdif_lock, flags);
	cmd.flags = MMC_RSP_NONE;
	cmd.arg = buff[0] << 24;
	cmd.arg += buff[1] << 16;
	cmd.arg += buff[2] << 8;
	cmd.arg += buff[3];
	cmd.opcode = opcode;
	spin_unlock_irqrestore(&sdif_lock, flags);
	mmc_wait_for_cmd(dt10sd_host, &cmd, 3);
}

/*==============================================================================*/
/*  Func:   _TP_BusOut                                                          */
/*  Desc:   Called by Test Point                                                */
/*==============================================================================*/
void _TP_BusOut(unsigned int addr, unsigned int data)
{
	unsigned char buff[4];
	struct mmc_command cmd;

#if DT_ADD_TIME_INFO
	struct timespec now;
	u32 time;
#endif
#if DT_ADD_CPU_INFO
	unsigned int cpu_id;
#endif

#if DT_ADD_TIME_INFO
	if (init == 0) {
		getnstimeofday(&now);
		start_time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	}
	getnstimeofday(&now);
	time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	time -= start_time;
	addr |= 0x40000;
#endif

#if DT_ADD_CPU_INFO
	cpu_id = smp_processor_id();
	data &= 0x3FFF;
	data |= (cpu_id << 14);
#endif

	/* Send SD CMD */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0))
	__mmc_claim_host(dt10sd_host, NULL);
#else
	if (mmc_try_claim_host(dt10sd_host) == 0) {
		return;
	}
#endif
	memset(&cmd, 0, sizeof(struct mmc_command));

	buff[0] = data >> 8;
	buff[1] = data;
	buff[2] = addr >> 12;
	buff[3] = addr >> 4;
	dt10_sd_send_cmd(buff, NML_TP_CMD);

#if DT_ADD_TIME_INFO
	buff[0] = time;
	buff[1] = time >> 8;
	buff[2] = time >> 16;
	buff[3] = time >> 24;
	dt10_sd_send_cmd(buff, TIME_INFO_CMD);
#endif

	mmc_release_host(dt10sd_host);
}
EXPORT_SYMBOL(_TP_BusOut);

/*==============================================================================*/
/*  Func:   _TP_MemoryOutput                                                    */
/*  Desc:   Called by Variable TestPoint                                        */
/*==============================================================================*/
void _TP_MemoryOutput(unsigned int addr, unsigned int data, void *value, unsigned int size)
{
	unsigned char buff[4];
	struct mmc_command cmd;
	unsigned char *p = (unsigned char *)value;
	int i, j, data_count;
#if DT_ADD_TIME_INFO
	struct timespec now;
	u32 time;
#endif
#if DT_ADD_CPU_INFO
	unsigned int cpu_id;
#endif

#if DT_ADD_TIME_INFO
	if (init == 0) {
		getnstimeofday(&now);
		start_time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	}
	getnstimeofday(&now);
	time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	time -= start_time;
	addr |= 0x40000;
#endif

#if DT_ADD_CPU_INFO
	cpu_id = smp_processor_id();
	data &= 0x3FFF;
	data |= (cpu_id << 14);
#endif

	if( size >= 256 ) size = 256;

	/* Send SD CMD */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0))
	__mmc_claim_host(dt10sd_host, NULL);
#else
	if (mmc_try_claim_host(dt10sd_host) == 0) {
		return;
	}
#endif
	memset(&cmd, 0, sizeof(struct mmc_command));

	buff[0] = data >> 8;
	buff[1] = data;
	buff[2] = addr >> 12;
	buff[3] = addr >> 4;
	dt10_sd_send_cmd(buff, VAL_TP_CMD);

#if DT_ADD_TIME_INFO
	buff[0] = time;
	buff[1] = time >> 8;
	buff[2] = time >> 16;
	buff[3] = time >> 24;
	dt10_sd_send_cmd(buff, TIME_INFO_CMD);
#endif

	buff[0] = 0x00;
	buff[1] = 0x00;
	buff[2] = (size+3)/4;
	buff[3] = size;
	dt10_sd_send_cmd(buff, VAL_SIZE_CMD);
	
	data_count = 0;
	for (i = 0; i < ((size+3)/4); i++) {
		for (j = 0; j < 4; j++, p++) {
			if (data_count < size) {
				buff[j] = *p;
				data_count++;
			} else {
				buff[j] = 0x00;
			}
		}
		dt10_sd_send_cmd(buff, VAL_DAT_CMD);
	}
	mmc_release_host(dt10sd_host);
}
EXPORT_SYMBOL(_TP_MemoryOutput);

/*==============================================================================*/
/*  Func:   _TP_EventTrigger32                                                  */
/*  Desc:   Called by Event Trigger 32bit TestPoint                             */
/*==============================================================================*/
void _TP_EventTrigger32(unsigned int addr, unsigned int data, unsigned int event_id)
{
	unsigned char buff[4];
	struct mmc_command cmd;
#if DT_ADD_TIME_INFO
	struct timespec now;
	u32 time;
#endif
#if DT_ADD_CPU_INFO
	unsigned int cpu_id;
#endif

#if DT_ADD_TIME_INFO
	if (init == 0) {
		getnstimeofday(&now);
		start_time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	}
	getnstimeofday(&now);
	time = (u32)(now.tv_sec*1000*1000 + now.tv_nsec/1000);
	time -= start_time;
	addr |= 0x40000;
#endif

#if DT_ADD_CPU_INFO
	cpu_id = smp_processor_id();
	data &= 0x3FFFF;
	data |= (cpu_id << 14);
#endif

	addr |= 0x80000;

	/* Send SD CMD */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 13, 0))
	__mmc_claim_host(dt10sd_host, NULL);
#else
	if (mmc_try_claim_host(dt10sd_host) == 0) {
		return;
	}
#endif
	memset(&cmd, 0, sizeof(struct mmc_command));

	buff[0] = data >> 8;
	buff[1] = data;
	buff[2] = addr >> 12;
	buff[3] = addr >> 4;
	dt10_sd_send_cmd(buff, NML_TP_CMD);

#if DT_ADD_TIME_INFO
	buff[0] = time;
	buff[1] = time >> 8;
	buff[2] = time >> 16;
	buff[3] = time >> 24;
	dt10_sd_send_cmd(buff, TIME_INFO_CMD);
#endif

	buff[0] = event_id;
	buff[1] = event_id >> 8;
	buff[2] = event_id >> 16;
	buff[3] = event_id >> 24;
	dt10_sd_send_cmd(buff, EVENT_ID_CMD);

	mmc_release_host(dt10sd_host);
}
EXPORT_SYMBOL(_TP_EventTrigger32);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static ssize_t proc_write(struct file *filp, const char __user *buf, size_t len, loff_t *data)
#else
static int proc_write(struct file *filp, const char *buf, unsigned long len, void *data)
#endif
{
	struct DT10_DATA dt10_data;

	if (copy_from_user(&dt10_data, buf, len)) {
		return -EFAULT;
	}

	switch (dt10_data.tp_func) {
		case TP_BusOut:
			_TP_BusOut(dt10_data.addr, dt10_data.dat);
			break;
		case TP_MemoryOutput:
			_TP_MemoryOutput(dt10_data.addr, dt10_data.dat, dt10_data.value, dt10_data.size);
			break;
		default:
			break;
	}

	return len;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
static const struct file_operations dt_proc_fops = {
	.write = proc_write,
};
#endif

static int __init dt10_sdif_init(void)
{
	struct proc_dir_entry* entry;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0))
	entry = proc_create(PROC_NAME, 666, NULL, &dt_proc_fops);
	if ( !entry ) {
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
#else
	entry = create_proc_entry(PROC_NAME, 0666, NULL);
	if ( entry ) {
		entry->write_proc = proc_write;
	} else {
		printk( KERN_ERR "proc entry fail\n" );
		return -EBUSY;
	}
#endif
	init = 1;
	printk( KERN_INFO "dt10sdif load OK\n" );
	return 0;
}

static void __exit dt10_sdif_exit(void)
{
	remove_proc_entry(PROC_NAME, NULL);
	printk( KERN_INFO "dt10sdif unload\n" );
}

module_init( dt10_sdif_init );
module_exit( dt10_sdif_exit );

MODULE_DESCRIPTION("dt10sdif");
MODULE_LICENSE("GPL2");
